import { RectangularStructureParams } from "./rectangular_structure_params";
import { Structure } from "./structure";
import { StructureCem88d } from "./structure_cem88d";
import { StructureCem88v } from "./structure_cem88v";
import { StructureCunge80 } from "./structure_cunge80";
import { StructureKivi } from "./structure_kivi";
import { StructureKiviParams } from "./structure_kivi_params";
import { StructureOrificeFree } from "./structure_orifice_free";
import { StructureOrificeSubmerged } from "./structure_orifice_submerged";
import { LoiDebit, StructureProperties, StructureType } from "./structure_props";
// tslint:disable-next-line:max-line-length
import { StructureTriangularTruncWeirFree, TriangularTruncStructureParams } from "./structure_triangular_trunc_weir_free";
import { StructureTriangularWeirFree, TriangularStructureParams } from "./structure_triangular_weir_free";
import { StructureWeirFree } from "./structure_weir_free";
import { StructureWeirSubmergedLarinier } from "./structure_weir_submerged_larinier";

export function CreateStructure(structureType: StructureType, loiDebit: LoiDebit, dbg: boolean = false): Structure {
    const oCd: {[s: string]: number} = {SeuilR: 0.4, VanneR: 0.6, SeuilT: 1.36};
    const rectStructPrms: RectangularStructureParams = new RectangularStructureParams(
        0,  // Q
        100,        // ZDV
        102,        // Z1
        101.5,      // Z2
        2,          // L
        oCd.SeuilR        // Cd pour un seuil rectangulaire
        // W = Infinity par défaut pour un seuil
    );

    // Control of validity of structure type and definition of parameters
    switch (structureType) {
        case StructureType.VanneRectangulaire:
            rectStructPrms.W.v = 0.5;
            rectStructPrms.Cd.v = oCd.VanneR; // Cd pour une vanne rectangulaire
            break;
        case StructureType.SeuilRectangulaire:
        case StructureType.SeuilTriangulaire:
        case StructureType.SeuilTriangulaireTrunc:
            break;

        default:
            throw new Error(`type de structure ${StructureType[structureType]} non pris en charge`);
    }

    // Control validity of the couple Structure Type / Discharge equation
    if (!(StructureProperties.isCompatibleValues(structureType, loiDebit))) {
        throw new Error(
            `La loi de débit ${LoiDebit[loiDebit]} n'est pas admissible pour le type ${StructureType[structureType]}`
        );
    }

    // Instanciation of the equation
    switch (loiDebit) {
        case LoiDebit.Cem88d:
            rectStructPrms.Cd.v = oCd.SeuilR; // Cd Cem88d qu'on soit en seuil ou vanne
            return new StructureCem88d(rectStructPrms, dbg);

        case LoiDebit.Cem88v:
            rectStructPrms.Cd.v = oCd.VanneR; // Cd Cem88d qu'on soit en seuil ou vanne
            return new StructureCem88v(rectStructPrms, dbg);

        case LoiDebit.Cunge80:
            rectStructPrms.Cd.v = oCd.VanneR; // Cd Cunge80 qu'on soit en seuil ou vanne
            return new StructureCunge80(rectStructPrms, dbg);

        case LoiDebit.OrificeFree:
            return new StructureOrificeFree(rectStructPrms, dbg);

        case LoiDebit.OrificeSubmerged:
            return new StructureOrificeSubmerged(rectStructPrms, dbg);

        case LoiDebit.WeirFree:
            return new StructureWeirFree(rectStructPrms, dbg);

        case LoiDebit.WeirSubmergedLarinier:
            return new StructureWeirSubmergedLarinier(rectStructPrms, dbg);

        case LoiDebit.KIVI:
            const structKiviPrm: StructureKiviParams = new StructureKiviParams(
                8.516,      // Q
                101,        // ZDV
                103,        // Z1
                102,        // Z2
                2,          // L
                oCd.SeuilR, // alpha
                0.001,      // béta
                100);       // ZRAM : cote Radier Amont
            return new StructureKivi(structKiviPrm, dbg);
            case LoiDebit.TriangularWeirFree:
            const structTriangPrms: TriangularStructureParams = new TriangularStructureParams(
                0,          // Q
                100,        // ZDV
                102,        // Z1
                45,         // Alpha2
                oCd.SeuilT  // Cd pour un seuil triangulaire
                // W = Infinity par défaut pour un seuil
            );
            return new StructureTriangularWeirFree(structTriangPrms, dbg);
            case LoiDebit.TriangularTruncWeirFree:
            const structTriTruncPrms: TriangularTruncStructureParams = new TriangularTruncStructureParams(
                0,          // Q
                100.1,        // ZDV
                102,        // Z1
                0.9,        // BT
                101,        // ZT
                oCd.SeuilT  // Cd pour un seuil triangulaire
                // W = Infinity par défaut pour un seuil
            );
            return new StructureTriangularTruncWeirFree(structTriTruncPrms, dbg);

        default:
            throw new Error(`type de LoiDebit ${LoiDebit[loiDebit]} non pris en charge`);
    }
}
